package;

import haxe.io.Input;

using StringTools;

/**
   List of keys on the keyboard that are reported by an escape sequence in ansi
**/
enum KeyboardKey {
   BasicKey(character:Int, ctrl:Bool, alt:Bool, shift:Bool);
   Escape; // ~
   ArrowUp; // ~[A
   ArrowDown; // ~[B
   ArrowRight; // ~[C
   ArrowLeft; // ~[D
   End; // ~[F
   Home; // ~[H
   Insert; // ~[2~
   Delete; // ~[3~
   PageUp; // ~[5~
   PageDown; // ~[6~
   CtrlC; // 3
   Other(code:Int);
}

enum abstract ControlCharacter(Int) from Int to Int {
   var ARROW_UP = 'A'.code;
   var ARROW_DOWN = 'B'.code;
   var ARROW_RIGHT = 'C'.code;
   var ARROW_LEFT = 'D'.code;
   var KEY_HOME = 'H'.code;
   var KEY_END = 'F'.code;
   var PAGE_UP = '5'.code;
   var PAGE_DOWN = '6'.code;
   var MOUSE_CLICK = 'M'.code;
}

enum abstract Colors(String) from String to String {
   var BLACK = '\x1b[30m';
   var RED = '\x1b[31m';
   var GREEN = '\x1b[32m';
   var YELLOW = '\x1b[33m';
   var BLUE = '\x1b[34m';
   var MAGENTA = '\x1b[35m';
   var CYAN = '\x1b[36m';
   var GRAY = '\x1b[37m';
   var GREY = '\x1b[37m';

   var BLACK_BG = '\x1b[40m';
   var RED_BG = '\x1b[41m';
   var GREEN_BG = '\x1b[42m';
   var YELLOW_BG = '\x1b[43m';
   var BLUE_BG = '\x1b[44m';
   var MAGENTA_BG = '\x1b[45m';
   var CYAN_BG = '\x1b[46m';
   var GRAY_BG = '\x1b[47m';
   var GREY_BG = '\x1b[47m';

   var RESET = '\x1b[0m';

   var BOLD = '\x1b[1m';
   var FAINT = '\x1b[2m';
   var NORMAL_WEIGHT = '\x1b[22m';

   var UNDERLINED = '\x1b[4m';
   var NOT_UNDERLINED = '\x1b[24m';

   var INVERTED = '\x1b[7m';
   var NOT_INVERTED = '\x1b[27m';
}

enum BasicAnsiColors {
   BLACK;
   RED;
   GREEN;
   YELLOW;
   BLUE;
   MAGENTA;
   CYAN;
   WHITE;
}

enum AnsiColor {
   BasicColor(color:BasicAnsiColors, bright:Bool);
}

typedef CursorPosition = {
   var row:Int;
   var column:Int;
}

typedef TerminalSize = {
   var width:Int;
   var height:Int;
}

class ANSI {
   /**
      If the alternate screen buffer is active or not.
   **/
   public static var alternateScreen(default, set):Bool = false;

   static function set_alternateScreen(value:Bool):Bool {
      printAnsi('?1049${value ? 'h' : 'l'}');
      return alternateScreen = value;
   }

   public static var reportMouseClicks(default, set):Bool = false;

   static function set_reportMouseClicks(value:Bool):Bool {
      printAnsi('?9${value ? 'h' : 'l'}');
      return reportMouseClicks = value;
   }

   public static var alternateScrollMode(default, set):Bool = false;

   static function set_alternateScrollMode(value:Bool):Bool {
      printAnsi('1007${value ? 'h' : 'l'}');
      return alternateScrollMode = value;
   }

   /**
      The current position of the text cursor in the terminal
   **/
   public static var cursorPosition(get, set):CursorPosition;

   static function set_cursorPosition(pos:CursorPosition):CursorPosition {
      printAnsi('${pos.row};${pos.column}H');
      return pos;
   }

   static function get_cursorPosition():CursorPosition {
      printAnsi('6n');
      var response = '';
      while (!response.endsWith('R')) {
         response += String.fromCharCode(Sys.getChar(false));
      }
      var row = response.substring(2, response.indexOf(';'));
      var column = response.substring(response.indexOf(';') + 1, response.length - 1);
      return {row: Std.parseInt(row), column: Std.parseInt(column)};
   }

   /**
      The size of the terminal window
   **/
   public static var terminalSize(get, never):TerminalSize;

   static var lastTerminalSize:TerminalSize = {width: -1, height: -1};
   static var lastTerminalSizeGet:Float = -1.0;

   static function get_terminalSize() {
      if (Sys.cpuTime() - lastTerminalSizeGet < 0.2) {
         return lastTerminalSize;
      }

      lastTerminalSizeGet = Sys.cpuTime();

      printAnsi('s');
      printAnsi('999;999f');
      var size = cursorPosition;
      printAnsi('u');
      return lastTerminalSize = {width: size.column, height: size.row};
   }

   static var colorInts:Map<BasicAnsiColors, Int> = [
      BLACK => 30,
      RED => 31,
      GREEN => 32,
      YELLOW => 33,
      BLUE => 34,
      MAGENTA => 35,
      CYAN => 36,
      WHITE => 37
   ];

   /**
      print an ansi escape sequence. automatically adds `\x1b[` to the start
   **/
   static function printAnsi(ansi:String):Void {
      Sys.stdout().writeString('\x1b[$ansi');
   }

   /**
      strips ansi formatting from a string. useful to get the printed length of a string.
      this does not correctly handle *every* escape sequence, only formatting ones
   **/
   public static function stripFormatting(str:String):String {
      var newStr = '';

      var idx:Int = 0;

      function peek(offset:Int = 0):Null<Int> {
         return str.charCodeAt(idx + offset);
      }

      function next():Null<Int> {
         return str.charCodeAt(idx++);
      }

      while (idx < str.length - 1) {
         var c = next();

         switch c {
            case 27: // escape char
               var c2 = next();

               if (c2 != '['.code) { // not the only CSI character, but i dont care about the others for now
                  newStr += String.fromCharCode(c2);
                  continue;
               }

               var c3 = next();
               while (c3 != 'm'.code && c3 != null) {}
            default:
               newStr += String.fromCharCode(c);
         }
      }

      return str;
   }

   static var stdin:Input = Sys.stdin();

   public static function readChar():Null<Int> {
      return stdin.readByte();
   }

   public static function readCharacter(echo:Bool = false):KeyboardKey {
      var char = Sys.getChar(false);

      switch char {
         case 3:
            return CtrlC;
         case 27: // ANSI escape sequence start
            var second = Sys.getChar(false);

            switch second {
               case '['.code:
                  var third = Sys.getChar(false);

                  switch third {
                     case 'A'.code:
                        return ArrowUp;
                     case 'B'.code:
                        return ArrowDown;
                     case 'C'.code:
                        return ArrowRight;
                     case 'D'.code:
                        return ArrowLeft;
                     case 'F'.code:
                        return End;
                     case 'H'.code:
                        return Home;
                     case '2'.code:
                        Sys.getChar(false); // skip the trailing ~
                        return Other(0);
                  }
            }
      }
      return Other(0);
   }

   public static function prettyPrintableCharFromCode(code:Int):String {
      var raw = false;
      var printable = switch code {
         case 0: '^@';
         case 1: '^A';
         case 2: '^B';
         case 3: '^C';
         case 4: '^D';
         case 5: '^E';
         case 6: '^F';
         case 7: '\\a';
         case 8: '\\b';
         case 9: '\\t';
         case 10: '\\n';
         case 11: '\\v';
         case 12: '\\f';
         case 13: '\\r';
         case 14: '^N';
         case 15: '^O';
         case 16: '^P';
         case 17: '^Q';
         case 18: '^R';
         case 19: '^S';
         case 20: '^T';
         case 21: '^U';
         case 22: '^V';
         case 23: '^W';
         case 24: '^K';
         case 25: '^Y';
         case 26: '^Z';
         case 27: '^[';
         case 28: '^\\';
         case 29: '^]';
         case 30: '^^';
         case 31: '^_';
         case 127: '^?';
         default:
            raw = true;
            String.fromCharCode(code);
      };
      if (!raw)
         printable = '${ANSI.Colors.BLUE}${printable}${ANSI.Colors.RESET}';
      return printable;
   }

   public static function createFormattingString(?color:AnsiColor, ?bold:Bool, ?underline:Bool):String {
      var str = '\x1b[0';

      switch color {
         case null: // nothing!
         case BasicColor(color, bright):
            var col = colorInts.get(color);
            if (bright)
               col += 10;
            str += ';${col}';
      }

      str += 'm';
      return str;
   }

   /**
      Clears the screen
   **/
   public static function clearScreen(mode:ClearMode = EraseAll):Void {
      printAnsi('${mode}J');
   }
}

enum abstract ClearMode(String) from String to String {
   var EraseBelow = '0';
   var EraseAbove = '1';
   var EraseAll = '2';
   var EraseSaved = '3'; // xterm
}
