import interfaces.DebugInterface;
import interfaces.ListEntry;
import interfaces.SelectorInterface;
import haxe.zip.Entry;
import haxe.zip.Reader;
import haxe.io.BytesInput;
import haxe.Json;
import haxe.DynamicAccess;
import sys.io.File;
import haxe.io.Path;
import sys.FileSystem;

using StringTools;

enum abstract GodotMirror(String) from String to String {
   var GITHUB = 'https://github.com/godotengine/godot/releases/download';
   var GITHUB_BUILDS = 'https://github.com/godotengine/godot-builds/releases/download';
   var TUXFAMILY = 'https://downloads.tuxfamily.org/godotengine';
   var DEFAULT = '{default}'; // GITHUB for stable releases, GITHUB_BUILDS for previews
}

enum abstract GodotPlatform(String) to String {
   var Android = 'android.apk';
   var Linux32 = 'linux.32';
   var Linux64 = 'linux.64';
   var LinuxServer64 = 'linux_server.64';
   var LinuxServerHeadless64 = 'linux_server.headless.64';
   var Macos = 'macos.universal';
   var Web = 'web';
   var Windows32 = 'windows.32';
   var Windows64 = 'windows.64';
   var ExportTemplates = 'export_templates';
   var Auto = '{auto}'; // Linux64 on linux, Macos on mac, Windows64 on windows
}

enum GodotVersionChoice {
   EnvOverrideFile(path:String); // GODOT_EXECUTABLE envvar
   EnvOverrideVersion(version:String); // GODOT_VERSION envvar
   LocalOverrideFile(path:String); // unused for now
   LocalOverrideVersion(version:String); // .godot_version in cwd
   GlobalVersion(version:String); // .godot_version in program path
   None; // no version
}

typedef VersionDownloadConfig = {
   var templates:String;
   var editor:DynamicAccess<String>;
   var ?extras:{aar_library:String};
   var ?mono:VersionDownloadConfig;
}

var downloadConfigs:DynamicAccess<VersionDownloadConfig> = Json.parse(Macros.fileContent('./downloadConfigs.json'));
var host(get, never):GodotPlatform;
var _host:Null<GodotPlatform>;

function get_host():GodotPlatform {
   if (_host != null)
      return _host;

   var envVar = Sys.getEnv('GODOT_MANAGER_PLATFORM');
   if (envVar != null) {
      switch (envVar.toLowerCase()) {
         case 'android.apk' | 'linux.64' | 'linux.32' | 'macos.universal' | 'windows.64' | 'windows.32' | 'web' | 'linux_server.headless.64' |
            'linux_server.64': // pass raw platforms right through
            return _host = cast(envVar, GodotPlatform);
         case 'android':
            return _host = Android;
         case 'linux64' | 'linux':
            return _host = Linux64;
         case 'linux32':
            return _host = Linux32;
         case 'macos':
            return _host = Macos;
         case 'windows64' | 'windows' | 'win64' | 'win':
            return _host = Windows64;
         case 'windows32' | 'win32':
            return _host = Windows32;
         case 'linux_server.headless' | 'linux.headless' | 'linux_headless' | 'linux.headless.64' | 'linux_headless.64':
            return _host = LinuxServerHeadless64;
         case 'linux_server':
            return _host = LinuxServer64;
         case 'html' | 'html5':
            return _host = Web;
         case 'export' | 'templates':
            return _host = ExportTemplates;
         default:
            // default
      }
   }

   return _host = switch (Sys.systemName()) {
      case 'Windows':
         Windows64;
      case 'Linux' | 'BSD':
         Linux64;
      case 'Mac':
         Macos;
      default:
         Linux64; // how
   }
}

inline var GODOT_VERSION_FILE_NAME:String = '.godot_version';
var silent = false;

function main() {
   var args = Sys.args();

   if (args[0] == '-h') { // support for older verisons of godot-tools VSCode extension
      silent = true;
      // trace('help');
   }

   var targetVersion = getTargetVersion();

   if (args.contains('--version') && args[0] != 'run') {
      log('Version Folder: ${getVersionsFolder()}');
      log('Target Version: ${Std.string(targetVersion)}');
      log('Target Version Output: ');
      runGodot(targetVersion, ['--version']);
      return;
   }

   log('Godot Manager ${Macros.versionString()}-${Macros.commitHash().substr(0, 8)}');

   var cmd = args.shift().toLowerCase();

   switch cmd {
      #if debug
      case 'fixasb':
         ANSI.alternateScreen = false;
      #end
      case 'switch':
         if (targetVersion.match(EnvOverrideFile(_)) || targetVersion.match(EnvOverrideVersion(_))) {
            log('Warning: EnvVar override is active! Will override version file.');
         }
         pickVersion(args.contains('-local'), args.contains('-debug'));
      case 'download':
         downloadVersion(args.shift());
      case 'debug':
         new DebugInterface().show();
      case 'download_export_templates':
         // TODO
      case 'run':
         runGodot(targetVersion, args);
      default:
         runGodot(targetVersion, Sys.args());
   }
}

function runGodot(version:GodotVersionChoice, args:Array<String>, exit:Bool = true):Int {
   if (version.match(None)) {
      log('No Godot version has been configured. Use `godot switch` to select one, or `godot download` to download one.');
      if (exit)
         Sys.exit(1);
      return 1;
   }

   var path = resolveGodotVersionPath(version);
   if (!FileSystem.exists(path)) {
      log('Cannot find file for target Godot version: ${path}');
      if (exit)
         Sys.exit(1);
      return 1;
   }

   var exitCode = Sys.command(path, args);
   if (exit)
      Sys.exit(exitCode);

   return exitCode;
}

function log(text:String):Void {
   if (silent)
      return;
   Sys.println(text);
}

function exec(ansi:String):Void {
   if (silent)
      return;
   Sys.stdout().writeString('\033[$ansi');
   // Sys.print('\033[${ansi}');
}

typedef GodotVersionData = { // ex 4.2-beta1, 4.1.2
   var name:String; // 4.2, 4.1.2
   var flavor:String; // beta1, stable
   var isStable:Bool;
   var ?platform:String;
}

function parseVersionName(version:String):GodotVersionData {
   if (version.startsWith('v'))
      version = version.substr(1); // remove leading `v`, eg `v4.1-stable` -> `4.1-stable`

   if (!version.contains('-')) // release versions dont always include -stable in the name
      return {name: version, flavor: 'stable', isStable: true};

   var parts = version.split('-');
   return {name: parts[0], flavor: parts[1], isStable: parts[1] == 'stable'};
}

function versionDataToString(data:GodotVersionData):String {
   return 'v${data.name}-${data.flavor}';
}

function getDownloadData(version:String, platform:GodotPlatform = Auto, mono:Bool = false, mirror:GodotMirror = DEFAULT):Null<{url:String, file:String}> {
   var versionData = parseVersionName(version);
   var versionParts = versionData.name.split('.');
   var slug = '';
   var monoSlug = '';

   if (!downloadConfigs.exists(versionParts[0])) {
      log('Invalid version');
      return null;
   }

   var defaults = downloadConfigs.get(versionParts[0]);
   if (mono) {
      if (defaults.mono == null) {
         log('Godot ${version} does not support mono!');
         return null;
      }

      monoSlug = '/mono';
      defaults = defaults.mono;
   }

   if (platform == Auto) {
      platform = host;
   }

   if (!defaults.editor.exists(platform)) {
      log('Godot v${versionData.name}${mono ? ' mono' : ''} does not have a published ${platform} build!');
      return null;
   }

   slug = defaults.editor.get(platform);

   var file = 'Godot_v${versionData.name}-${versionData.flavor}_${slug}';

   if (versionParts[0] == '1' || (versionParts[0] == '2' && versionParts[1] == '0'))
      file = 'Godot_v${versionData.name}_${versionData.flavor}_${slug}';

   if (mirror == DEFAULT)
      mirror = (version == '1.0' || !versionData.isStable) ? GITHUB_BUILDS : GITHUB;

   var url = '$mirror/${versionData.name}-${versionData.flavor}/$file';

   if (mirror == TUXFAMILY) {
      url = '$mirror/${versionData.name}/${versionData.flavor}$monoSlug/$file';
      if (versionData.isStable)
         url = '$mirror/${versionData.name}$monoSlug/$file';
   }

   return {url: url, file: file.split('.zip')[0]};
}

function getVersionDataFromFileName(filename:String):Null<GodotVersionData> {
   if (!~/Godot_v\d\.\d(?>\.\d)?[-_][a-z]+\d?_.+/.match(filename)) {
      return null;
   }
   var split = filename.split('_');
   var version = split[1];
   split.shift();
   if (!((version.startsWith('v1') || version.startsWith('v2.0'))))
      split = split[0].split('-');

   return {name: split[0].substring(1), flavor: split[1], isStable: split[1] == 'stable'};
}

function getMirrorToUse():GodotMirror {
   var envVar = Sys.getEnv('GODOT_MANAGER_DOWNLOAD_MIRROR');
   if (envVar == null) {
      return DEFAULT;
   }

   switch (envVar.toLowerCase()) {
      case 'tuxfamily' | 'tux':
         return TUXFAMILY;
      case 'github' | 'gh':
         return GITHUB;
      case 'github_builds' | 'gh_b' | 'ghb':
         return GITHUB_BUILDS;
      default:
         return DEFAULT;
   }
}

function download(version:String, platform:GodotPlatform = Auto, mono:Bool = false, mirror:GodotMirror = DEFAULT) {
   var download = getDownloadData(version, platform, mono, mirror);

   if (download == null) {
      log('Unable to download v${version}');
      Sys.exit(1);
      return;
   }

   var path = Path.normalize(getVersionsFolder() + '/temp/');
   if (!FileSystem.exists(path))
      FileSystem.createDirectory(path);

   var zipFilePath = Path.normalize(path + '/download.zip');

   var request:FileDownloader = new FileDownloader(download.url, zipFilePath);
   request.download();

   var fileBytes = File.getBytes(zipFilePath);
   var bytesInput = new BytesInput(fileBytes);
   var reader = new Reader(bytesInput);
   var entries:List<Entry> = reader.read();
   for (_entry in entries) {
      // trace(entry);
      var data = Reader.unzip(_entry);
      if (_entry.fileName.substring(_entry.fileName.lastIndexOf('/') + 1) == '' && _entry.data.toString() == '') {
         FileSystem.createDirectory(Path.join([path, _entry.fileName]));
         continue;
      }
      var f = File.write(Path.join([path, _entry.fileName]), true);
      f.write(data);
      f.close();
   }
   // FileSystem.rename(Path.normalize(path + '/' + executable), Path.normalize(path + '/../$executable'));
   // deleteRecursive(Path.normalize(path + '/../temp'));
}

function downloadVersion(?version:String):Void {
   if (version == null) {
      log('Please enter a Godot version (eg 4.1)');
      Sys.exit(1);
      return;
   }

   var download = getDownloadData(version, Auto, false, getMirrorToUse());

   if (download == null) {
      log('Unable to download v${version}');
      Sys.exit(1);
      return;
   }

   var url = download.url;

   log('Downloading Godot ${version} from ${url}...');
   var path = Path.normalize(getVersionsFolder() + '/temp/');
   var executable = download.file;

   if (!FileSystem.exists(path))
      FileSystem.createDirectory(path);

   var zipFilePath = Path.normalize(path + '/download.zip');

   // download the file
   var request:FileDownloader = new FileDownloader(url, zipFilePath);
   request.download();

   var fileBytes = File.getBytes(zipFilePath);
   var bytesInput = new BytesInput(fileBytes);
   var reader = new Reader(bytesInput);
   var entries:List<Entry> = reader.read();
   for (_entry in entries) {
      var data = Reader.unzip(_entry);
      if (_entry.fileName.substring(_entry.fileName.lastIndexOf('/') + 1) == '' && _entry.data.toString() == '') {
         FileSystem.createDirectory(Path.join([path, _entry.fileName]));
         continue;
      }
      var f = File.write(Path.join([path, _entry.fileName]), true);
      f.write(data);
      f.close();
   }
   FileSystem.rename(Path.normalize(path + '/' + executable), Path.normalize(path + '/../$executable'));
   deleteRecursive(Path.normalize(path + '/../temp'));

   if (['Linux', 'BSD'].contains(Sys.systemName()) && host.contains('linux')) {
      Sys.command('chmod', ['+x', Path.normalize(path + '/../$executable')]);
   }

   log('Complete!');
}

function deleteRecursive(path:String) {
   var content = FileSystem.readDirectory(path);
   for (file in content) {
      if (FileSystem.isDirectory(Path.join([path, file]))) {
         deleteRecursive(Path.join([path, file]));
         continue;
      }
      FileSystem.deleteFile(Path.join([path, file]));
   }
   FileSystem.deleteDirectory(path);
}

function pickVersion(local:Bool = false, debug:Bool = false):Void {
   var path = getVersionsFolder();
   var versionList:Array<String> = FileSystem.readDirectory('$path');

   if (versionList.length == 0) {
      log('No Godot versions have been added to the versions folder. Use godot download <version> or add binaries to the versions folder in the godot-manager folder.');
      return;
   }

   versionList.sort((a, b) -> {
      a = a.toUpperCase();
      b = b.toUpperCase();

      if (a < b)
         return -1;
      if (a > b)
         return 1;
      return 0;
   });

   var namedVersionList = versionList.map((version) -> {
      var data = getVersionDataFromFileName(version);
      if (data == null)
         return version;

      return 'Godot v${data.name}${data.isStable ? '' : '-${data.flavor}'}';
   });

   var selector = new SelectorInterface();
   var entries:Array<ListEntry> = cast versionList.map(version -> {
      var name = version;
      var data = getVersionDataFromFileName(version);

      if (data != null)
         name = 'Godot v${data.name}${data.isStable ? '' : '-${data.flavor}'}';

      return new VersionListEntry(selector, name, version);
   });
   entries[0].selected();
   selector.entries = entries;
   var result = selector.select();
   switch result {
      case Selected(entry):
         if (!(entry is VersionListEntry)) {
            log('what');
            return;
         }

         var versionEntry:VersionListEntry = cast entry;

         File.saveContent(getVersionFilePath(local), versionEntry.data);

         log('Switched to ${versionEntry.name}');
      case Aborted:
         log('Version unchanged');
   }
}

function getVersionsFolder():String {
   var env = Sys.getEnv('GODOT_VERSION_FOLDER');
   if (env != null) {
      if (FileSystem.exists(Path.normalize(env))) {
         return Path.normalize(env);
      }
   }

   return Path.normalize(Sys.programPath() + '/../versions/') + '/';
}

function getVersionFilePath(local:Bool = true):String {
   var path = Sys.programPath() + '/..';

   if (local)
      path = Sys.getCwd();

   return Path.normalize(Path.join([path, GODOT_VERSION_FILE_NAME]));
}

function getTargetVersion(silent:Bool = false, requireFile:Bool = true):GodotVersionChoice {
   inline function out(text:String) {
      if (!silent)
         log(text);
   }

   var envFile = Sys.getEnv('GODOT_EXECUTABLE');
   if (envFile != null) {
      if (FileSystem.exists(envFile))
         return EnvOverrideFile(envFile);
      out('GODOT_EXECUTABLE doesnt exist!');
      return None;
   }

   var envVersion = Sys.getEnv('GODOT_VERSION');
   if (envVersion != null) {
      var path = getVersionsFolder() + envVersion;
      if (FileSystem.exists(path) || !requireFile)
         return EnvOverrideVersion(envVersion);

      out('GODOT_VERSION doesnt exist!');
      return None;
   }

   var localVersionFilePath = Path.normalize(Sys.getCwd() + '/${GODOT_VERSION_FILE_NAME}');
   if (FileSystem.exists(localVersionFilePath)) {
      var localVersion = File.getContent(localVersionFilePath);
      if (localVersion.length > 1) {
         var path = getVersionsFolder() + StringTools.trim(localVersion);
         if (FileSystem.exists(path) || !requireFile)
            return LocalOverrideVersion(localVersion);

         out('Local override version doesnt exist! $path');
         return None;
      }

      out('invalid content in local version file');
      return None;
   }

   var globalVersionFilePath = Path.normalize(Sys.programPath() + '/../${GODOT_VERSION_FILE_NAME}');
   if (FileSystem.exists(globalVersionFilePath)) {
      var globalVersion = File.getContent(globalVersionFilePath);
      if (globalVersion.length > 1) {
         var path = getVersionsFolder() + StringTools.trim(globalVersion);
         if (FileSystem.exists(path) || !requireFile)
            return GlobalVersion(globalVersion);

         out('Global version doesnt exist!');
      }
   }

   return None;
}

function resolveGodotVersionPath(version:GodotVersionChoice):String {
   var versionFolder = getVersionsFolder();

   switch version {
      case EnvOverrideFile(path):
         return path;
      case EnvOverrideVersion(version):
         return '$versionFolder$version';
      case LocalOverrideFile(path): // unused for now
         return path;
      case LocalOverrideVersion(version):
         return '$versionFolder$version';
      case GlobalVersion(version):
         return '$versionFolder$version';
      case None:
         return 'None';
   }
}
