package interfaces;

import ANSI.BasicAnsiColors;
import interfaces.SelectorInterface;

using StringTools;

abstract class ListEntry {
   var list:ListInterface;
   var isSelected:Bool;

   public function new(list:ListInterface) {
      this.list = list;
   };

   abstract public function print():Void;

   public function selected():Void {
      this.isSelected = true;
   };

   public function deselected():Void {
      this.isSelected = false;
   };

   public function keyLeft():Void {};

   public function keyRight():Void {};

   public function keyEnter():Void {};

   public function click(x:Int):Void {};
}

class BasicListEntry extends ListEntry {
   public var name:String;

   public function new(list:ListInterface, name:String) {
      super(list);
      this.name = name;
   }

   public function print():Void {
      Sys.print('${this.isSelected ? '>' : ' '} ${name}');
   }
}

typedef ListButton = {
   var display:String;
   var color:String;
   var selected_background:String;
   var selected_color:String;
   var ?data:String;
}

class ButtonedListEntry extends BasicListEntry {
   static final SELECTED_OPEN_BRACKET = '{';
   static final SELECTED_CLOSED_BRACKET = '}';
   static final UNSELECTED_OPEN_BRACKET = '[';
   static final UNSELECTED_CLOSED_BRACKET = ']';

   public var currentButton:Int = -1;

   public var data:String;

   var buttons:Array<ListButton>;
   var buttonCallback:(ListButton) -> Void;
   // buttonsType is used when many entries have the same options, it will keep the same button selected when switching list entries
   var buttonsType:Int;

   public function new(list:ListInterface, name:String, buttons:Array<ListButton>, data:String, buttonCallback:(ListButton) -> Void, buttonsType:Int = -1) {
      super(list, name);
      this.data = data;
      this.buttons = buttons;
      this.buttonCallback = buttonCallback;
      this.buttonsType = buttonsType;
   }

   public override function print():Void {
      var display = '';

      if (isSelected)
         display += '${ANSI.Colors.CYAN}${ANSI.Colors.BOLD}';

      display += name;
      display += ANSI.Colors.RESET;

      // var currentRawDisplay = ANSI.stripFormatting(display);
      var len = name.length;

      var visibleTextLength = len;

      var buttonsDisplay = '';

      for (idx => button in buttons) {
         visibleTextLength += button.display.length + 3; // text length + open bracket + closed bracket + spacer'
         buttonsDisplay += ' ${button.color}';
         buttonsDisplay += idx == currentButton ? SELECTED_OPEN_BRACKET : UNSELECTED_OPEN_BRACKET;
         if (idx == currentButton) {
            buttonsDisplay += button.selected_background;
            buttonsDisplay += button.selected_color;
         }
         buttonsDisplay += button.display;
         if (idx == currentButton) {
            buttonsDisplay += ANSI.Colors.RESET;
            buttonsDisplay += button.color;
         }
         buttonsDisplay += idx == currentButton ? SELECTED_CLOSED_BRACKET : UNSELECTED_CLOSED_BRACKET;
         buttonsDisplay += ANSI.Colors.RESET;
      }

      while (len < Std.int(Math.min(50, ANSI.terminalSize.width - visibleTextLength))) {
         display += ' ';
         len++;
      }

      Sys.print(display + buttonsDisplay);
   }

   function characterButtonIndex():Array<Int> {
      var mainArray = [];

      for (char in name) {
         mainArray.push(-1);
      }

      var len = name.length;
      var visibleTextLength = len;
      var buttonsArray = [];

      for (idx => button in buttons) {
         visibleTextLength += button.display.length + 3; // text length + open bracket + closed bracket + spacer

         buttonsArray.push(-1);
         buttonsArray.push(-1);
         for (char in button.display) {
            buttonsArray.push(idx);
         }
         buttonsArray.push(-1);
      }

      while (len < Std.int(Math.min(50, ANSI.terminalSize.width - visibleTextLength))) {
         mainArray.push(-1);
         len++;
      }

      return mainArray.concat(buttonsArray);
   }

   public override function deselected():Void {
      super.deselected();
      currentButton = -1;
   }

   public override function selected():Void {
      super.selected();
      if (this.buttonsType >= 0 && list.currentlySelected is ButtonedListEntry) {
         var previouslySelected:ButtonedListEntry = cast list.currentlySelected;

         if (previouslySelected.buttonsType == this.buttonsType) {
            this.currentButton = previouslySelected.currentButton;
            return;
         }
      }
      this.currentButton = 0;
   }

   public override function keyLeft():Void {
      super.keyLeft();
      currentButton--;
      if (currentButton < 0)
         currentButton = 0;
   }

   public override function keyRight():Void {
      super.keyRight();
      currentButton++;
      if (currentButton > buttons.length - 1)
         currentButton = buttons.length - 1;
   }

   public override function keyEnter():Void {
      buttonPress(buttons[currentButton]);
   }

   public override function click(x:Int):Void {
      var buttonIndexes = characterButtonIndex();

      if (buttonIndexes.length + 1 < x) {
         return; // click is out of bounds
      }

      var clickedIndex = buttonIndexes[x];

      if (clickedIndex < 0 || clickedIndex >= buttons.length) {
         return; // invalid click
      }

      // if we already have the clicked button selected, press the button
      if (currentButton == clickedIndex) {
         buttonPress(buttons[currentButton]);
         return;
      }

      // otherwise, select the button
      currentButton = clickedIndex;
   }

   function buttonPress(button:ListButton):Void {
      if (buttonCallback == null)
         return;

      buttonCallback(button);
   }
}

class VersionListEntry extends ButtonedListEntry {
   var selector:SelectorInterface;
   var deleteConfirmVisible(default, set):Bool = false;

   function set_deleteConfirmVisible(value:Bool):Bool {
      buttons[1].display = value ? 'confirm' : 'delete';
      return deleteConfirmVisible = value;
   }

   public function new(list:SelectorInterface, name:String, value:String) {
      super(list, name, [
         {
            display: 'select',
            color: ANSI.Colors.GREEN,
            selected_color: ANSI.Colors.BLACK,
            selected_background: ANSI.Colors.GREEN_BG,
         },
         {
            display: 'delete',
            color: ANSI.Colors.RED,
            selected_color: ANSI.Colors.BLACK,
            selected_background: ANSI.Colors.RED_BG,
         }
      ], value, buttonPress, 1);
      this.selector = list;
   }

   override function buttonPress(button:ListButton):Void {
      switch button.display {
         case 'select':
            selector.madeSelection = true;
            selector.hide();
         case 'delete':
            if (!deleteConfirmVisible) {
               deleteConfirmVisible = true;
               return;
            }

         // actually delete
         default:
            // invalid button, but just ignore it
      }
   }

   override function deselected():Void {
      super.deselected();
      deleteConfirmVisible = false;
   }
}
